﻿; ------------------------------------------------------------------------------------
; Easy Network Manager
; PB 4.5+ OS: Windows/Linux/MacOS Architecture: x86/x64
; ------------------------------------------------------------------------------------

; ------------------------------------------------------------------------------------
; Features:
;
; - vollständig Threadsicher
; - vollständiges Callback Managemant
; - einfach Multi-Server & Multi-Connections Verwaltung
; - einfacher Interface Zugriff
; - Client & Server Verwaltung
; - Client #PB_NetworkEvent_Disconnect
; 
; ------------------------------------------------------------------------------------


; Befehls-Beschreibung:
XIncludeFile "ENM_CommandDescription.pbi"

EnableExplicit

; Threadsafe muss aktiviert sein
CompilerIf #PB_Compiler_Thread = #Null
  CompilerError "Enable Thread-Safe!"
CompilerEndIf

XIncludeFile "SafeNetwork.pbi"
XIncludeFile "JobQueue_MultiThread_LE.pbi"

Prototype _ENM_Event_CB(Server, Client)

Enumeration 1 
  #ENM_Server = 2
  #ENM_Client = 1
  #ENM_Full = #ENM_Server|#ENM_Client
EndEnumeration

Enumeration 1
  #ENM_Event_Connect
  #ENM_Event_DisConnect
  #ENM_Event_Data
  #ENM_Event_File
EndEnumeration

Structure ENM_CallBacks
  *Connect._ENM_Event_CB
  *Disconnect._ENM_Event_CB
  *Receive._ENM_Event_CB
  *ReceiveFile._ENM_Event_CB
EndStructure

Structure _ENM_ClientData
  ID.i
  XData.i
EndStructure

; Strukturen
Structure _ENM_Server
  ID.i ; Server ID
  Port.l ; Server Port
  Type.l ; Server Type
  JQ.i ; Event Queue
  RecFile.l ; Datei wird entfangen
  RecData.l ; Daten werden entfangen
  Map Clients._ENM_ClientData() ; Clients am Server
  *CB.ENM_CallBacks ; Callbacks
EndStructure

Structure _ENM_Clients
  ID.i
  JQ.i
  RecFile.l
  RecData.l
  XData.i
  *CB.ENM_CallBacks
EndStructure

Structure _ENM_Interface_VTable
  *AddServer
  *RemoveServer
  *OpenConnection
  *CloseConnection
  *SetCallbacks
  *GetConnectionData
  *SetConnectionData
  *GetServerClientData
  *SetServerClientData
EndStructure


Structure ENM_Data_Server
  EventProzessor.i
  Stop.i
  GMutex.i
  Count.i
  Map Servers._ENM_Server()
EndStructure

Structure ENM_Data_Client
  EventProzessor.i
  Stop.i
  GMutex.i
  Count.i
  Map Clients._ENM_Clients()
EndStructure

Structure ENM_Data
  *VTable._ENM_Interface_VTable
  Type.l
  GBCollector.i
  Server.ENM_Data_Server
  Client.ENM_Data_Client
EndStructure

Structure _ENM_Event
  *Data.ENM_Data
  *CB._ENM_Event_CB
  *Con._ENM_Server
  Client.i
  Type.i
EndStructure

Interface _ENM_Interface
  AddServer(Port.l, Type.l = #PB_Network_TCP)
  RemoveServer(Server.i)
  OpenConnection(ServerName$, Port.l, Type = #PB_Network_TCP)
  CloseConnection(Connection.i, Server = #Null)
  SetCallbacks(Connection_Server.i, *CB.ENM_Callbacks)
  GetConnectionData(Connection)
  SetConnectionData(Connection, XData.i)
  GetServerClientData(Server, Client)
  SetServerClientData(Server, Client, XData.i)
EndInterface

Global ENM._ENM_Interface

; Interne Prozeduren


; JobQues beenden
Procedure __ENM_QuitJQ(JQ)
  JQ_WaitUntilFinish(JQ)
  JQ_DeInit(JQ)
EndProcedure
Macro _ENM_QuitJQ(_ENM, _JQ)
  JQ_Add(_ENM#\GBCollector, @__ENM_QuitJQ(), _JQ)
EndMacro

; Client - Daten laden
Procedure _ENM_GetServerClientData(*ENM.ENM_Data, Server, Client)
  Protected XData.i
  
  ; Server Initialisiert?
  If Not *ENM\Type & #ENM_Server
    ProcedureReturn #Null
  EndIf
  
  
  With *ENM\Server
    LockMutex(\GMutex)
    If \Servers(Str(Server))\ID ; Gültiger Server?
      If \Servers()\Clients(Str(Client))\ID ; Gültiger Client?
        XData = \Servers()\Clients()\XData
      Else ; Sonst Client-Entry löschen
        DeleteMapElement(\Servers()\Clients())
      EndIf
      
    Else ; Sonst Server-Entry löschen
      DeleteMapElement(\Servers())
    EndIf
    
    UnlockMutex(\GMutex)
  EndWith
  
  ProcedureReturn XData
EndProcedure

; Client - Daten speichern
Procedure _ENM_SetServerClientData(*ENM.ENM_Data, Server, Client, XData.i)
  
  ; Server initalisiert?
  If Not *ENM\Type & #ENM_Server
    ProcedureReturn #Null
  EndIf
  
  With *ENM\Server
    LockMutex(\GMutex)
    If \Servers(Str(Server))\ID ; Gültiger Server?
      If \Servers()\Clients(Str(Client))\ID ; Gültiger Client?
        \Servers()\Clients()\XData = XData
      Else
        DeleteMapElement(\Servers()\Clients())
      EndIf
      
    Else ; Sonst Server-Entry löschen
      DeleteMapElement(\Servers())
    EndIf
    
    UnlockMutex(\GMutex)
  EndWith
  
EndProcedure

; Server hinzufügen
Procedure _ENM_AddServer(*ENM.ENM_Data, Port.l, Type.l)
  
  ; Server initialisiert?
  If Not *ENM\Type & #ENM_Server
    ProcedureReturn #Null
  EndIf
  
  ; Server erstellen
  Protected Server = CreateNetworkServer(#PB_Any, Port, Type)
  
  With *ENM\Server
    If Server
      LockMutex(\GMutex)
      \Count + 1
      \Servers(Str(Server))\ID = Server
      \Servers()\Port = Port
      \Servers()\Type = Type
      ; JobQueue starten
      \Servers()\JQ = JQ_Init(1)
      UnlockMutex(\GMutex)
    EndIf
  EndWith
  
  ProcedureReturn Server
EndProcedure

; Server entfernen
Procedure _ENM_RemoveServer(*ENM.ENM_Data, Server.i)
  
  ; Server initialisiert?
  If Not *ENM\Type & #ENM_Server
    ProcedureReturn #Null
  EndIf
  
  With *ENM\Server
    LockMutex(\GMutex)  
    
    If \Servers(Str(Server))\ID 
      \Count-1
      ; Server beenden
      CloseNetworkServer(Server)
      ; JobQueue beenden
      _ENM_QuitJQ(*ENM, \Servers()\JQ)
    EndIf
    ; Server löschen
    DeleteMapElement(\Servers())
    
    UnlockMutex(\GMutex)
  EndWith
EndProcedure

; Xtra Daten lesen
Procedure _ENM_GetConnectionData(*ENM.ENM_Data, Connection)
  Protected XData.i
  
  ; Client initialisiert?
  If Not *ENM\Type & #ENM_Client
    ProcedureReturn #Null
  EndIf
  
  With *ENM\Client
    LockMutex(\GMutex)
    If \Clients(Str(Connection))\ID
      XData = \Clients()\XData
    Else
      DeleteMapElement(\Clients())
    EndIf
    UnlockMutex(\GMutex)
  EndWith
  
  ProcedureReturn XData
EndProcedure

; Xtra Daten setzen
Procedure _ENM_SetConnectionData(*ENM.ENM_Data, Connection, XData.i)

  ; Client initialisiert?
  If Not *ENM\Type & #ENM_Client
    ProcedureReturn #Null
  EndIf
  
  With *ENM\Client
    LockMutex(\GMutex)
    If \Clients(Str(Connection))\ID
      \Clients()\XData = XData
    Else
      DeleteMapElement(\Clients())
    EndIf
    UnlockMutex(\GMutex)
  EndWith
EndProcedure

Procedure _ENM_OpenConnection(*ENM.ENM_Data, ServerName$, Port.l, Type = #PB_Network_TCP)
  Protected Con
  
  ; Client initialisiert?
  If Not *ENM\Type & #ENM_Client
    ProcedureReturn #Null
  EndIf
  
  With *ENM\Client
    ; Connection erstellen und registrieren
    Con = OpenNetworkConnection(ServerName$, Port, Type)
    If Con ; Connection ok?
      LockMutex(\GMutex)
      \Count + 1
      \Clients(Str(Con))\ID = Con
      \Clients()\JQ = JQ_Init(1)
      UnlockMutex(\GMutex)
    EndIf
  EndWith

  ProcedureReturn Con
EndProcedure

Procedure _ENM_CloseConnection(*ENM.ENM_Data, Connection.i, Server = #Null)
  
  If Server ; Server Connection beenden
    ; Server initialisiert?
    If Not *ENM\Type & #ENM_Server
      ProcedureReturn #Null
    EndIf
    
    With *ENM\Server
      LockMutex(\GMutex)
      If \Servers(Str(Server))\ID ; Gültiger Server?
        If \Servers()\Clients(Str(Connection))\ID ; Gültiger Client?
          CloseNetworkConnection(\Servers()\Clients()\ID) ; Connection beenden
        EndIf
        DeleteMapElement(\Servers()\Clients())
      Else ; Sonst Server-Entry löschen
        DeleteMapElement(\Servers())
      EndIf
      
      UnlockMutex(\GMutex)
    EndWith
    
  Else ; Client Connection beenden
    
    ; Client initialisiert?
    If Not *ENM\Type & #ENM_Client
      ProcedureReturn #Null
    EndIf
    
    With *ENM\Client
      LockMutex(\GMutex)
      If \Clients(Str(Connection))\ID
        \Count - 1
        ; JobQueue beenden, Connection schließen
        CloseNetworkConnection(Connection)
        _ENM_QuitJQ(*ENM, \Clients()\JQ)
      EndIf
      DeleteMapElement(\Clients())
      
      UnlockMutex(\GMutex)
    EndWith
    
  EndIf

EndProcedure

; Callbacks einstellen
Procedure _ENM_SetCallbacks(*ENM.ENM_Data, Server_Con.i, *CB.ENM_Callbacks)
  
  With *ENM\Server
    LockMutex(\GMutex)  
    
    If *ENM\Type & #ENM_Server
      If \Servers(Str(Server_Con))\ID
        ; Callbacks einstellen
        \Servers()\CB = *CB
      Else
        DeleteMapElement(\Servers())
      EndIf
    EndIf
    UnlockMutex(\GMutex)
  EndWith
  
  With *ENM\Client
    If *ENM\Type & #ENM_Client
      If \Clients(Str(Server_Con))\ID
        ; Callbacks einstellen
        \Clients()\CB = *CB
      Else
        DeleteMapElement(\Clients())
      EndIf
    EndIf
    UnlockMutex(\GMutex)
  EndWith
    
  
EndProcedure

; zwischenebene für Eventprozessoren
Procedure _ENM_EventMiddleLayer(*ENME._ENM_Event)
  With *ENME
    \CB(\Con\ID, \Client)
    
    Select \Type
      Case #ENM_Event_File
        \Con\RecFile = #False
        
      Case #ENM_Event_Data
        \Con\RecData = #False
        
    EndSelect
    
    FreeMemory(*ENME)
  EndWith
EndProcedure


Procedure _ENM_SendClientEvent(*ENM.ENM_Data, *Client._ENM_Clients, Client, *CB, Type)
  Protected *ENMEvent._ENM_Event
  
  If *CB
    ; Event erstellen
    *ENMEvent = AllocateMemory(SizeOf(_ENM_Event))
    With *ENMEvent
      \CB = *CB
      \Data = *ENM
      \Con = *Client
      \Client = Client
      \Type = Type
    EndWith
    
    ; Auf Event-Stack legen
    JQ_Add(*Client\JQ, @_ENM_EventMiddleLayer(), *ENMEvent)
  EndIf
EndProcedure

; Event absetzen
Procedure _ENM_SendServerEvent(*ENM.ENM_Data, *Server._ENM_Server, Client, *CB, Type)
  Protected *ENMEvent._ENM_Event
  
  If *CB
    ; Event erstellen
    *ENMEvent = AllocateMemory(SizeOf(_ENM_Event))
    With *ENMEvent
      \CB = *CB
      \Data = *ENM
      \Con = *Server
      \Client = Client
      \Type = Type
    EndWith
    
    ; Auf Event-Stack legen
    JQ_Add(*Server\JQ, @_ENM_EventMiddleLayer(), *ENMEvent)
  EndIf
EndProcedure

; Events abfragen und absenden
Procedure _ENM_ServerEventProcessor(*ENM.ENM_Data)
  Protected Event, Client, Server, *ServerD._ENM_Server
  
  With *ENM\Server
    
    Repeat
      If \Stop ; StopSignal
        SignalSemaphore(\Stop)
        ProcedureReturn #True
      EndIf
      
      Event = NetworkServerEvent()
      
      If Event ; Netzwerk Event da
        ; Server abfragen
        Server = EventServer()
        ; Events abfragen und verrabeiten
        Client = EventClient()
        
        ; Callbacks laden
        LockMutex(\GMutex)  
        If \Servers(Str(Server))\ID
          *ServerD = @\Servers()
        Else
          DeleteMapElement(\Servers())
        EndIf
        UnlockMutex(\GMutex)
        
        ; Events verarbeiten und absenden
        Select Event 
          Case #PB_NetworkEvent_Data
            If *ServerD\CB And Not *ServerD\RecFile
              *ServerD\RecFile = #True
              _ENM_SendServerEvent(*ENM, *ServerD, Client, *ServerD\CB\Receive, #ENM_Event_Data)
            EndIf
            
          Case #PB_NetworkEvent_File
            If *ServerD\CB And Not *ServerD\RecData
              *ServerD\RecData = #True
              _ENM_SendServerEvent(*ENM, *ServerD, Client, *ServerD\CB\ReceiveFile, #ENM_Event_File)
            EndIf
            
          Case #PB_NetworkEvent_Connect
            ; Client registrieren und ConTime speichern 
            LockMutex(\GMutex)
            \Servers(Str(Server))\Clients(Str(Client))\ID = Client
            UnlockMutex(\GMutex)
            
            If *ServerD\CB
              _ENM_SendServerEvent(*ENM, *ServerD, Client, *ServerD\CB\Connect, #ENM_Event_Connect)
            EndIf
            
          Case #PB_NetworkEvent_Disconnect
            
            If *ServerD\CB
              _ENM_SendServerEvent(*ENM, *ServerD, Client, *ServerD\CB\Disconnect, #ENM_Event_DisConnect)
            EndIf
            
            ; Client austragen
            LockMutex(\GMutex)
            DeleteMapElement(\Servers(Str(Server))\Clients(), Str(Client))
            UnlockMutex(\GMutex)
        EndSelect
        
      Else
        Delay(5) ; Wir wollen ja nicht den kompletten Prozessor verheizen ...
      EndIf
      
    ForEver
    
  EndWith
EndProcedure


; Events abfragen und absenden
Procedure _ENM_ClientEventProcessor(*ENM.ENM_Data)
  Protected Event, Client, *ClientD._ENM_Clients
  
  With *ENM\Client
    
    Repeat
      If \Stop ; StopSignal
        SignalSemaphore(\Stop)
        ProcedureReturn #True
      EndIf
      
      If Not \Count
        Delay(5)
        Continue
      EndIf
      
      ForEach \Clients() ; Alle Clients durchgehen
        Client = \Clients()\ID
        Event = NetworkClientEvent(Client)
        
        If Event ; Netzwerk Event da
          
          ; Callbacks laden
          LockMutex(\GMutex)  
          *ClientD = @\Clients(Str(Client))
          UnlockMutex(\GMutex)
          
          ; Events verarbeiten und absenden
          Select Event 
            Case #PB_NetworkEvent_Data
              If *ClientD\CB And Not *ClientD\RecFile
                *ClientD\RecFile = #True
                _ENM_SendClientEvent(*ENM, *ClientD, Client, *ClientD\CB\Receive, #ENM_Event_Data)
              EndIf
              
            Case #PB_NetworkEvent_File
              If *ClientD\CB And Not *ClientD\RecData
                *ClientD\RecData = #True
                _ENM_SendClientEvent(*ENM, *ClientD, Client, *ClientD\CB\ReceiveFile, #ENM_Event_File)
              EndIf
              
            Case #PB_NetworkEvent_Disconnect
              
              If *ClientD\CB
                _ENM_SendClientEvent(*ENM, *ClientD, Client, *ClientD\CB\Disconnect, #ENM_Event_DisConnect)
              EndIf
              _ENM_CloseConnection(*ENM, Client)
              
          EndSelect
          
        Else
          Delay(5) ; Wir wollen ja nicht den kompletten Prozessor verheizen ...
        EndIf
      Next
      
    ForEver
    
  EndWith
EndProcedure



; Public

; ENM Initialisieren
Procedure ENM_Init(Type = #ENM_Full)
  Protected *ENM2.ENM_Data = AllocateMemory(SizeOf(ENM_Data))
  Protected *VTable = AllocateMemory(SizeOf(_ENM_Interface_VTable))
  Shared ENM
  
  InitializeStructure(*ENM2, ENM_Data)
  
  With *ENM2
    ; VTable anlegen
    \VTable = *VTable
    \VTable\AddServer = @_ENM_AddServer()
    \VTable\RemoveServer = @_ENM_RemoveServer()
    \VTable\OpenConnection = @_ENM_OpenConnection()
    \VTable\CloseConnection = @_ENM_CloseConnection()
    \VTable\SetCallbacks = @_ENM_SetCallbacks()
    \VTable\GetConnectionData = @_ENM_GetConnectionData()
    \VTable\SetConnectionData = @_ENM_SetConnectionData()
    \VTable\GetServerClientData = @_ENM_GetServerClientData()
    \VTable\SetServerClientData = @_ENM_SetServerClientData()
    
    ; Server starten
    \Type = Type
    If \Type & #ENM_Server
      \Server\GMutex = CreateMutex()
      \Server\EventProzessor = CreateThread(@_ENM_ServerEventProcessor(), *ENM2)
    EndIf
    
    ; Client erstellen
    If \Type & #ENM_Client
      \Client\GMutex = CreateMutex()
      \Client\EventProzessor = CreateThread(@_ENM_ClientEventProcessor(), *ENM2)
    EndIf
    
    ; Garbage Collector
    \GBCollector = JQ_Init(1)
  EndWith
  
  PokeI(@ENM, *ENM2)
  
EndProcedure

Procedure ENM_DeInit()
  Protected *ENM2.ENM_Data = PeekI(@ENM)
  
  With *ENM2
    If \Type & #ENM_Server
      ; Alle Server beenden
      If \Server\Count
        ForEach \Server\Servers()
          _ENM_RemoveServer(*ENM2, \Server\Servers()\ID)
        Next
      EndIf
      ; Event - Verwaltung beenden
      \Server\Stop = CreateSemaphore()
      WaitSemaphore(\Server\Stop)
      FreeSemaphore(\Server\Stop)
      ; Mutex freigeben
      FreeMutex(\Server\GMutex)
    EndIf
    
    ; Clients beenden
    If \Type & #ENM_Client
      ; Alle Clients beenden
      
      If \Client\Count
        ForEach \Client\Clients()
          _ENM_CloseConnection(*ENM2, \Client\Clients()\ID)
        Next
      EndIf
      
      ; Client - Verwaltung beenden
      \Client\Stop = CreateSemaphore()
      WaitSemaphore(\Client\Stop)
      FreeSemaphore(\Client\Stop)
      ; Mutex freigeben
      FreeMutex(\Client\GMutex)
    EndIf
    
    ; Garbage freigeben
    JQ_WaitUntilFinish(\GBCollector)
    JQ_DeInit(\GBCollector)
    
    PokeI(@ENM, #Null)
    FreeMemory(*ENM2)
  EndWith
  
EndProcedure

; ------------------------------


; IDE Options = PureBasic 4.51 (Windows - x64)
; CursorPosition = 685
; FirstLine = 649
; Folding = ----
; EnableXP
; CPU = 1